@@ -1,4 +1,4 @@
-/*	Copyright (C) 1998, 1999 Free Software Foundation, Inc.
+/*	Copyright (C) 1998, 1999, 2000 Free Software Foundation, Inc.
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -134,17 +134,17 @@ static SCM scm_goops_the_unbound_value;
 static SCM scm_goops_lookup_closure;
 
 /* Some classes are defined in libguile/objects.c. */
-static SCM scm_class_top, scm_class_object, scm_class_class;
-static SCM scm_class_entity, scm_class_entity_with_setter;
-static SCM scm_class_generic, scm_class_generic_with_setter, scm_class_method;
-static SCM scm_class_simple_method, scm_class_accessor;
-static SCM scm_class_procedure_class;
-static SCM scm_class_operator_class, scm_class_operator_with_setter_class;
-static SCM scm_class_entity_class;
-static SCM scm_class_number, scm_class_list;
-static SCM scm_class_keyword;
-static SCM scm_class_port, scm_class_input_output_port;
-static SCM scm_class_input_port, scm_class_output_port;
+SCM scm_class_top, scm_class_object, scm_class_class;
+SCM scm_class_entity, scm_class_entity_with_setter;
+SCM scm_class_generic, scm_class_generic_with_setter, scm_class_method;
+SCM scm_class_simple_method, scm_class_accessor;
+SCM scm_class_procedure_class;
+SCM scm_class_operator_class, scm_class_operator_with_setter_class;
+SCM scm_class_entity_class;
+SCM scm_class_number, scm_class_list;
+SCM scm_class_keyword;
+SCM scm_class_port, scm_class_input_output_port;
+SCM scm_class_input_port, scm_class_output_port;
 SCM scm_class_foreign_class, scm_class_foreign_object;
 SCM scm_class_foreign_slot;
 SCM scm_class_self, scm_class_protected;
@@ -166,6 +166,102 @@ scm_goops_version ()
   return scm_makfrom0str (GOOPS_VERSION);
 }
 
+SCM_PROC (s_oldfmt, "oldfmt", 1, 0, 0, scm_oldfmt);
+
+char *
+scm_c_oldfmt (char *from, int n)
+{
+#ifdef HAVE_SCM_SIMPLE_FORMAT
+  return from;
+#else
+  static struct { int n; char *from; char *to; } *strings;
+  static int size = 0;
+  static int n_strings = 0;
+  char *to;
+  int i;
+  
+  for (i = 0; i < n_strings; ++i)
+    if (n == strings[i].n && strncmp (from, strings[i].from, n) == 0)
+      return strings[i].to;
+  
+  if (n_strings == size)
+    {
+      if (size == 0)
+	{
+	  size = 10;
+	  strings = scm_must_malloc (size * sizeof (*strings), s_oldfmt);
+	}
+      else
+	{
+	  int oldsize = size;
+	  size = 3 * oldsize / 2;
+	  strings = scm_must_realloc (strings,
+				      oldsize * sizeof (*strings),
+				      size * sizeof (*strings),
+				      s_oldfmt);
+	}
+    }
+
+  strings[n_strings].n = n;
+  strings[n_strings].from = strncpy (scm_must_malloc (n, s_oldfmt), from, n);
+  to = strings[n_strings].to = scm_must_malloc (n + 1, s_oldfmt);
+  n_strings++;
+
+  for (i = 0; i < n; ++i)
+    {
+      if (from[i] == '~' && ++i < n)
+	{
+	  if (from[i] == 'A')
+	    {
+	      to[i - 1] = '%';
+	      to[i] = 's';
+	    }
+	  else if (from[i] == 'S')
+	    {
+	      to[i - 1] = '%';
+	      to[i] = 'S';
+	    }
+	  else
+	    {
+	      to[i - 1] = '~';
+	      to[i] = from[i];
+	    }
+	  continue;
+	}
+      to[i] = from[i];
+    }
+  to[i] = '\0';
+  
+  return to;
+#endif
+}
+
+char *
+scm_c_oldfmt0 (char *s)
+{
+#ifdef HAVE_SCM_SIMPLE_FORMAT
+  return s;
+#else
+  return scm_c_oldfmt (s, strlen (s));
+#endif
+}
+
+SCM
+scm_oldfmt (SCM s)
+{
+#ifdef HAVE_SCM_SIMPLE_FORMAT
+  return s;
+#else
+  int n;
+  SCM_ASSERT (SCM_NIMP (s) && SCM_STRINGP (s), s, 1, s_oldfmt);
+  n = SCM_LENGTH (s);
+  return scm_return_first (scm_makfromstr (scm_c_oldfmt (SCM_ROCHARS (s), n),
+					   n,
+					   0),
+			   s);
+#endif
+}
+
 /******************************************************************************
  *
  * Compute-cpl
@@ -249,7 +345,9 @@ remove_duplicate_slots (SCM l, SCM res, SCM slots_already_seen)
 
   tmp = SCM_CAAR (l);
   if (!(SCM_NIMP (tmp) && SCM_SYMBOLP (tmp)))
-    scm_misc_error ("%compute-slots", "bad slot name %S", SCM_LIST1 (tmp));
+    scm_misc_error ("%compute-slots",
+		    scm_c_oldfmt0 ("bad slot name ~S"),
+		    SCM_LIST1 (tmp));
   
   if (SCM_NULLP (scm_sloppy_memq (tmp, slots_already_seen))) {
     res 	       = scm_cons (SCM_CAR (l), res);
@@ -350,7 +448,9 @@ scm_i_get_keyword (SCM key, SCM l, int len, SCM default_value, const char *subr)
   for (i = 0; i < len; i += 2)
     {
       if (!(SCM_NIMP (SCM_CAR (l)) && SCM_KEYWORDP (SCM_CAR (l))))
-	scm_misc_error (subr, "bad keyword: %S", SCM_LIST1 (SCM_CAR (l)));
+	scm_misc_error (subr,
+			scm_c_oldfmt0 ("bad keyword: ~S"),
+			SCM_LIST1 (SCM_CAR (l)));
       if (SCM_CAR (l) == key)
 	return SCM_CADR (l);
       l = SCM_CDDR (l);
@@ -365,10 +465,13 @@ scm_get_keyword (SCM key, SCM l, SCM default_value)
 {
   int len;
   SCM_ASSERT (SCM_NIMP (key) && SCM_KEYWORDP (key),
-	      key, "Bad keyword: %S", s_get_keyword);
+	      key,
+	      scm_c_oldfmt0 ("Bad keyword: ~S"),
+	      s_get_keyword);
   len = scm_ilength (l);
   SCM_ASSERT (len >= 0 && (len & 1) == 0, l,
-	      "Bad keyword-value list: %S", s_get_keyword);
+	      scm_c_oldfmt0 ("Bad keyword-value list: ~S"),
+	      s_get_keyword);
   return scm_i_get_keyword (key, l, len, default_value, s_get_keyword);
 }
 
@@ -409,7 +512,7 @@ scm_sys_initialize_object (SCM obj, SCM initargs)
 	  int n = scm_ilength (SCM_CDR (slot_name));
 	  if (n & 1) /* odd or -1 */
 	    scm_misc_error (s_sys_initialize_object,
-			    "class contains bogus slot definition: %S",
+			    scm_c_oldfmt0 ("class contains bogus slot definition: ~S"),
 			    SCM_LIST1 (slot_name));
 	  tmp 	= scm_i_get_keyword (k_init_keyword,
 				     SCM_CDR (slot_name),
@@ -422,7 +525,7 @@ scm_sys_initialize_object (SCM obj, SCM initargs)
 	      /* an initarg was provided for this slot */
 	      if (!(SCM_NIMP (tmp) && SCM_KEYWORDP (tmp)))
 		scm_misc_error (s_sys_initialize_object,
-				"initarg must be a keyword. It was %S",
+				scm_c_oldfmt0 ("initarg must be a keyword. It was ~S"),
 				SCM_LIST1 (tmp));
 	      slot_value = scm_i_get_keyword (tmp,
 					      initargs,
@@ -478,13 +581,13 @@ scm_sys_prep_layout_x (SCM class)
   nfields = SCM_SLOT (class, scm_si_nfields);
   if (!SCM_INUMP (nfields) || SCM_INUM (nfields) < 0)
     scm_misc_error (s_sys_prep_layout_x,
-		    "bad value in nfields slot: %S",
+		    scm_c_oldfmt0 ("bad value in nfields slot: ~S"),
 		    SCM_LIST1 (nfields));
   n = 2 * SCM_INUM (nfields);
   if (n < sizeof (SCM_CLASS_CLASS_LAYOUT) - 1
       && SCM_SUBCLASSP (class, scm_class_class))
     scm_misc_error (s_sys_prep_layout_x,
-		    "class object doesn't have enough fields: %S",
+		    scm_c_oldfmt0 ("class object doesn't have enough fields: ~S"),
 		    SCM_LIST1 (nfields));
   
   s  = n > 0 ? scm_must_malloc (n, s_sys_prep_layout_x) : 0;
@@ -559,9 +662,17 @@ scm_sys_inherit_magic_x (SCM class, SCM dsupers)
   else
     {
       int n = SCM_INUM (SCM_SLOT (class, scm_si_nfields));
+#if 0
+      /*
+       * We could avoid calling scm_must_malloc in the allocation code
+       * (in which case the following two lines are needed).  Instead
+       * we make 0-slot instances non-light, so that the light case
+       * can be handled without special cases.
+       */
       if (n == 0)
 	SCM_SET_CLASS_DESTRUCTOR (class, scm_struct_free_0);
-      else if (!(flags & SCM_CLASSF_METACLASS))
+#endif
+      if (n > 0 && !(flags & SCM_CLASSF_METACLASS))
 	{
 	  /* NOTE: The following depends on scm_struct_i_size. */
 	  flags |= SCM_STRUCTF_LIGHT + n * sizeof (SCM); /* use light representation */
@@ -582,7 +693,7 @@ prep_hashsets (SCM class)
 
   for (i = 0; i < 7; ++i)
     SCM_SLOT (class, scm_si_hashsets + i)
-      = scm_c_uniform32 (goops_rstate);
+      = SCM_PACK (scm_c_uniform32 (goops_rstate));
 }
 
 /******************************************************************************/
@@ -913,6 +1024,17 @@ scm_method_procedure (SCM obj)
   return scm_slot_ref (obj, Intern ("procedure"));
 }
 
+SCM_PROC (s_accessor_method_slot_definition, "accessor-method-slot-definition", 1, 0, 0, scm_accessor_method_slot_definition);
+
+SCM
+scm_accessor_method_slot_definition (SCM obj)
+{
+  SCM_ASSERT (SCM_NIMP (obj) && SCM_ACCESSORP (obj),
+	      obj, SCM_ARG1, s_method_procedure);
+  return scm_slot_ref (obj, Intern ("slot-definition"));
+}  
+
+
 /******************************************************************************
  *
  * S l o t   a c c e s s
@@ -1104,6 +1226,34 @@ test_slot_existence (SCM class, SCM obj, SCM slot_name)
 
 SCM_PROC1 (s_sys_logand, "%logand", scm_tc7_asubr, scm_sys_logand);
 
+static unsigned long
+scm_sloppy_num2ulong (SCM num, char *pos, const char *s_caller)
+{
+  unsigned long res;
+
+  if (SCM_INUMP (num))
+    {
+      if (SCM_INUM (num) < 0)
+	goto out_of_range;
+      res = SCM_INUM (num);
+      return res;
+    }
+  SCM_ASRTGO (SCM_NIMP (num), wrong_type_arg);
+  if (SCM_BIGP (num))
+    {
+      scm_sizet l;
+
+      res = 0;
+      for (l = SCM_NUMDIGS (num); l--;)
+	res = SCM_BIGUP (res) + SCM_BDIGITS (num)[l];
+      return res;
+    }
+ wrong_type_arg:
+  scm_wrong_type_arg (s_caller, (int) pos, num);
+ out_of_range:
+  scm_out_of_range (s_caller, num);
+}
+
 static SCM
 scm_sys_logand (SCM n1, SCM n2)
 {
@@ -1114,12 +1264,53 @@ scm_sys_logand (SCM n1, SCM n2)
       return n1;
     }
   {
-    unsigned long u1 = scm_num2ulong (n1, (char *) 1, s_sys_logand);
-    unsigned long u2 = scm_num2ulong (n2, (char *) 2, s_sys_logand);
+    unsigned long u1 = scm_sloppy_num2ulong (n1, (char *) 1, s_sys_logand);
+    unsigned long u2 = scm_sloppy_num2ulong (n2, (char *) 2, s_sys_logand);
     return scm_ulong2num (u1 & u2);
   }
 }
 
+/* These are needed in save.scm. */
+#ifndef HAVE_SCM_SHARED_ARRAY_ROOT
+
+SCM_PROC (s_shared_array_root, "shared-array-root", 1, 0, 0, scm_shared_array_root);
+
+SCM
+scm_shared_array_root (SCM ra)
+{
+  SCM_ASSERT (SCM_NIMP (ra) && SCM_ARRAYP (ra), ra, SCM_ARG1, s_shared_array_root);
+  return SCM_ARRAY_V (ra);
+}
+
+
+SCM_PROC (s_shared_array_offset, "shared-array-offset", 1, 0, 0, scm_shared_array_offset);
+
+SCM
+scm_shared_array_offset (SCM ra)
+{
+  SCM_ASSERT (SCM_NIMP (ra) && SCM_ARRAYP (ra), ra, SCM_ARG1, s_shared_array_offset);
+  return SCM_MAKINUM (SCM_ARRAY_BASE (ra));
+}
+
+
+SCM_PROC (s_shared_array_increments, "shared-array-increments", 1, 0, 0, scm_shared_array_increments);
+
+SCM
+scm_shared_array_increments (SCM ra)
+{
+  SCM res = SCM_EOL;
+  scm_sizet k;
+  scm_array_dim *s;
+  SCM_ASSERT (SCM_NIMP (ra) && SCM_ARRAYP (ra), ra, SCM_ARG1, s_shared_array_increments);
+  k = SCM_ARRAY_NDIM (ra);
+  s = SCM_ARRAY_DIMS (ra);
+  while (k--)
+    res = scm_cons (SCM_MAKINUM (s[k].inc), res);
+  return res;
+}
+
+#endif /* HAVE_SCM_SHARED_ARRAY_ROOT */
+
 		/* ======================================== */
 
 SCM_PROC (s_slot_ref_using_class, "slot-ref-using-class", 3, 0, 0, scm_slot_ref_using_class);
@@ -1277,8 +1468,8 @@ wrap_init (SCM class, SCM *m, int n)
     m[i] = SCM_GOOPS_UNBOUND;
 
   SCM_NEWCELL (z);
-  SCM_SETCDR (z, m);
-  SCM_SETCAR (z, (SCM) SCM_STRUCT_DATA (class) + scm_tc3_cons_gloc);
+  SCM_SETCDR (z, (SCM) m);
+  SCM_SETCAR (z, (scm_bits_t) SCM_STRUCT_DATA (class) | scm_tc3_cons_gloc);
 
   return z;
 }
@@ -1313,21 +1504,18 @@ scm_sys_allocate_instance (SCM class, SCM initargs)
     {
       m = (SCM *) scm_alloc_struct (n,
 				    scm_struct_entity_n_extra_words,
-				    "instance");
+				    "entity");
+      m[scm_struct_i_setter] = SCM_BOOL_F;
+      m[scm_struct_i_procedure] = SCM_BOOL_F;
       /* Generic functions */
       if (SCM_CLASS_FLAGS (class) & SCM_CLASSF_PURE_GENERIC)
 	{
 	  SCM gf = wrap_init (class, m, n);
-	  m[scm_struct_i_setter] = SCM_BOOL_F;
 	  clear_method_cache (gf);
 	  return gf;
 	}
       else
-	{
-	  m[scm_struct_i_setter] = SCM_BOOL_F;
-	  m[scm_struct_i_procedure] = SCM_BOOL_F;
-	  return wrap_init (class, m, n);
-	}
+	return wrap_init (class, m, n);
     }
   
   /* Class objects */
@@ -1354,7 +1542,7 @@ scm_sys_allocate_instance (SCM class, SCM initargs)
   {
     m = (SCM *) scm_alloc_struct (n,
 				  scm_struct_n_extra_words,
-				  "instance");
+				  "heavy instance");
     return wrap_init (class, m, n);
   }
 }
@@ -1466,6 +1654,7 @@ burnin (SCM o)
 static void
 go_to_hell (void *o)
 {
+  SCM obj = (SCM) o;
 #ifdef USE_THREADS
   scm_mutex_lock (&hell_mutex);
 #endif
@@ -1475,7 +1664,7 @@ go_to_hell (void *o)
       hell = scm_must_realloc (hell, hell_size, new_size, "hell");
       hell_size = new_size;
     }
-  hell[n_hell++] = SCM_INST (o);
+  hell[n_hell++] = SCM_INST (obj);
 #ifdef USE_THREADS
   scm_mutex_unlock (&hell_mutex);
 #endif
@@ -1651,7 +1840,7 @@ applicablep (SCM actual, SCM formal)
     }
     else 
       scm_misc_error (0,
-		      "Internal error in applicable: bad list %S",
+		      scm_c_oldfmt0 ("Internal error in applicable: bad list ~S"),
 		      SCM_LIST1 (actual));
   }
   return 0;
@@ -2064,7 +2253,7 @@ scm_find_method (SCM l)
   SCM_ASSERT (SCM_NIMP (gf) && GENERICP (gf), gf, SCM_ARG1, s_find_method);
   if (SCM_NULLP (SCM_SLOT (gf, scm_si_methods)))
     scm_misc_error (s_find_method,
-		    "no methods for generic %S",
+		    scm_c_oldfmt0 ("no methods for generic ~S"),
 		    SCM_LIST1 (gf));
 
   return scm_compute_applicable_methods (gf, l, len - 1, 1);
@@ -2119,30 +2308,37 @@ make_stdcls (SCM *var, char *name, SCM meta, SCM super, SCM slots)
    DEFVAR(tmp, *var);
 }
 
+
+SCM_KEYWORD (k_slot_definition, "slot-definition");
+
 static void
 create_standard_classes (void)
 {
-  SCM tmp0;
-  SCM tmp1 = SCM_LIST4 (Intern ("generic-function"), 
-			Intern ("specializers"), 
-			Intern ("procedure"),
-			Intern ("code-table"));
+  SCM slots;
+  SCM method_slots = SCM_LIST4 (Intern ("generic-function"), 
+				Intern ("specializers"), 
+				Intern ("procedure"),
+				Intern ("code-table"));
+  SCM amethod_slots = SCM_LIST1 (SCM_LIST3 (Intern ("slot-definition"),
+					    k_init_keyword,
+					    k_slot_definition));
 #ifdef USE_THREADS
-  SCM tmp2 = SCM_LIST1 (Intern ("make-mutex"));
+  SCM mutex_slot = SCM_LIST1 (Intern ("make-mutex"));
 #else
-  SCM tmp2 = SCM_BOOL_F;
+  SCM mutex_slot = SCM_BOOL_F;
 #endif
-  SCM tmp3 = SCM_LIST4 (Intern ("methods"),
-			SCM_LIST3 (Intern ("n-specialized"),
-				   k_init_value,
-				   SCM_INUM0),
-			SCM_LIST3 (Intern ("used-by"),
-				   k_init_value,
-				   SCM_BOOL_F),
-			SCM_LIST3 (Intern ("cache-mutex"),
-				   k_init_thunk,
-				   scm_closure (SCM_LIST2 (SCM_EOL, tmp2),
-						SCM_EOL)));
+  SCM gf_slots = SCM_LIST4 (Intern ("methods"),
+			    SCM_LIST3 (Intern ("n-specialized"),
+				       k_init_value,
+				       SCM_INUM0),
+			    SCM_LIST3 (Intern ("used-by"),
+				       k_init_value,
+				       SCM_BOOL_F),
+			    SCM_LIST3 (Intern ("cache-mutex"),
+				       k_init_thunk,
+				       scm_closure (SCM_LIST2 (SCM_EOL,
+							       mutex_slot),
+						    SCM_EOL)));
 
   /* Foreign class slot classes */
   make_stdcls (&scm_class_foreign_slot,	   "<foreign-slot>",
@@ -2176,11 +2372,11 @@ create_standard_classes (void)
 
   /* Continue initialization of class <class> */
   
-  tmp0 = build_class_class_slots ();
-  SCM_SLOT (scm_class_class, scm_si_direct_slots) = tmp0;
-  SCM_SLOT (scm_class_class, scm_si_slots) = tmp0;
+  slots = build_class_class_slots ();
+  SCM_SLOT (scm_class_class, scm_si_direct_slots) = slots;
+  SCM_SLOT (scm_class_class, scm_si_slots) = slots;
   SCM_SLOT (scm_class_class, scm_si_getters_n_setters)
-    = compute_getters_n_setters (tmp0);
+    = compute_getters_n_setters (slots);
   
   make_stdcls (&scm_class_foreign_class, "<foreign-class>",
 	       scm_class_class, scm_class_class,
@@ -2205,19 +2401,19 @@ create_standard_classes (void)
 	       "<operator-with-setter-class>",
 	       scm_class_class, scm_class_operator_class, SCM_EOL);
   make_stdcls (&scm_class_method,	   "<method>",
-	       scm_class_class, scm_class_object,	   tmp1);
+	       scm_class_class, scm_class_object,	   method_slots);
   make_stdcls (&scm_class_simple_method,   "<simple-method>",
 	       scm_class_class, scm_class_method,	   SCM_EOL);
   SCM_SET_CLASS_FLAGS (scm_class_simple_method, SCM_CLASSF_SIMPLE_METHOD);
   make_stdcls (&scm_class_accessor,	   "<accessor-method>",
-	       scm_class_class, scm_class_simple_method,   SCM_EOL);
+	       scm_class_class, scm_class_simple_method,   amethod_slots);
   SCM_SET_CLASS_FLAGS (scm_class_accessor, SCM_CLASSF_ACCESSOR_METHOD);
   make_stdcls (&scm_class_entity,	   "<entity>",
 	       scm_class_entity_class, scm_class_object,   SCM_EOL);
   make_stdcls (&scm_class_entity_with_setter, "<entity-with-setter>",
 	       scm_class_entity_class, scm_class_entity,   SCM_EOL);
   make_stdcls (&scm_class_generic,	   "<generic>",
-	       scm_class_entity_class, scm_class_entity,   tmp3);
+	       scm_class_entity_class, scm_class_entity,   gf_slots);
   SCM_SET_CLASS_FLAGS (scm_class_generic, SCM_CLASSF_PURE_GENERIC);
   make_stdcls (&scm_class_generic_with_setter, "<generic-with-setter>",
 	       scm_class_entity_class,
@@ -2331,8 +2527,14 @@ create_smob_classes (void)
   for (i = 0; i < 255; ++i)
     scm_smob_class[i] = 0;
 
+#ifdef scm_tc16_big
+  scm_smob_class[SCM_TC2SMOBNUM (scm_tc16_big)] = scm_class_integer;
+  scm_smob_class[SCM_TC2SMOBNUM (scm_tc16_real)] = scm_class_real;
+  scm_smob_class[SCM_TC2SMOBNUM (scm_tc16_complex)] = scm_class_complex;
+#else
   scm_smob_class[SCM_TC2SMOBNUM (scm_tc16_bigpos)] = scm_class_integer;
   scm_smob_class[SCM_TC2SMOBNUM (scm_tc16_bigneg)] = scm_class_integer;
+#endif
   scm_smob_class[SCM_TC2SMOBNUM (scm_tc16_keyword)] = scm_class_keyword;
   
   for (i = 0; i < scm_numsmob; ++i)
@@ -2526,8 +2728,8 @@ scm_wrap_object (SCM class, void *data)
 {
   SCM z;
   SCM_NEWCELL (z);
-  SCM_SETCDR (z, data);
-  SCM_SETCAR (z, SCM_CDR (class) + scm_tc3_cons_gloc);
+  SCM_SETCDR (z, (SCM) data);
+  SCM_SETCAR (z, SCM_UNPACK (SCM_CDR (class)) | scm_tc3_cons_gloc);
   return z;
 }
 
