@@ -2,7 +2,7 @@
 
 #ifndef GOOPSH
 #define GOOPSH
-/*	Copyright (C) 1998, 1999 Free Software Foundation, Inc.
+/*	Copyright (C) 1998, 1999, 2000 Free Software Foundation, Inc.
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -54,6 +54,11 @@
 #include <libguile.h>
 #include <goops-snarf.h>
 
+#ifndef SCM_PACK
+#define SCM_PACK(x) x
+#define SCM_UNPACK(x) x
+#endif
+
 /*
  * scm_class_class
  */
@@ -120,8 +125,9 @@ typedef struct scm_method_t {
 /* Also defined in libguuile/objects.c */
 #define SCM_CLASS_OF(x)        SCM_STRUCT_VTABLE (x)
 #define SCM_ACCESSORS_OF(x)    (SCM_STRUCT_VTABLE_DATA (x)[scm_si_getters_n_setters])
-#define SCM_NUMBER_OF_SLOTS(x) (SCM_STRUCT_DATA (x)[scm_struct_i_n_words] \
-			       - scm_struct_n_extra_words) \
+#define SCM_NUMBER_OF_SLOTS(x)\
+ (SCM_UNPACK (SCM_STRUCT_DATA (x)[scm_struct_i_n_words]) \
+  - scm_struct_n_extra_words) \
 
 #define SCM_INSTANCEP(x)       (SCM_STRUCTP (x) \
 			       && (SCM_INST_TYPE (x) & SCM_CLASSF_GOOPS))
@@ -168,6 +174,17 @@ typedef struct scm_method_t {
 #define scm_si_code_table	 3  /* offset of code. slot in a <method> */
 
 /* C interface */
+extern SCM scm_class_top, scm_class_object, scm_class_class;
+extern SCM scm_class_entity, scm_class_entity_with_setter;
+extern SCM scm_class_generic, scm_class_generic_with_setter, scm_class_method;
+extern SCM scm_class_simple_method, scm_class_accessor;
+extern SCM scm_class_procedure_class;
+extern SCM scm_class_operator_class, scm_class_operator_with_setter_class;
+extern SCM scm_class_entity_class;
+extern SCM scm_class_number, scm_class_list;
+extern SCM scm_class_keyword;
+extern SCM scm_class_port, scm_class_input_output_port;
+extern SCM scm_class_input_port, scm_class_output_port;
 extern SCM scm_class_foreign_class, scm_class_foreign_object;
 extern SCM scm_class_foreign_slot;
 extern SCM scm_class_self, scm_class_protected;
@@ -178,6 +195,9 @@ extern SCM scm_class_int, scm_class_float, scm_class_double;
 extern const char *scm_s_slot_set_x;
 
 SCM scm_goops_version (void);
+SCM scm_oldfmt (SCM);
+char *scm_c_oldfmt0 (char *);
+char *scm_c_oldfmt (char *, int n);
 void scm_load_goops (void);
 SCM scm_make_foreign_object (SCM class, SCM initargs);
 SCM scm_make_class (SCM meta, char *s_name, SCM supers, size_t size,
@@ -233,6 +253,7 @@ SCM scm_generic_function_methods (SCM obj);
 SCM scm_method_generic_function (SCM obj); 
 SCM scm_method_specializers (SCM obj); 
 SCM scm_method_procedure (SCM obj); 
+SCM scm_accessor_method_slot_definition (SCM obj); 
 SCM scm_sys_fast_slot_ref (SCM obj, SCM index); 
 SCM scm_sys_fast_slot_set_x (SCM obj, SCM index, SCM value); 
 SCM scm_slot_ref_using_class (SCM class, SCM obj, SCM slot_name); 
