@@ -1,6 +1,6 @@
 ;;; installed-scm-file
 
-;;;; 	Copyright (C) 1998, 1999 Free Software Foundation, Inc.
+;;;; 	Copyright (C) 1998, 1999, 2000 Free Software Foundation, Inc.
 ;;;; 
 ;;;; This program is free software; you can redistribute it and/or modify
 ;;;; it under the terms of the GNU General Public License as published by
@@ -75,7 +75,8 @@
     generic-function-name
     generic-function-methods method-generic-function method-specializers
     primitive-generic-generic enable-primitive-generic!
-    method-procedure slot-exists? make find-method get-keyword
+    method-procedure accessor-method-slot-definition
+    slot-exists? make find-method get-keyword
     %logand)
 
 
@@ -84,7 +85,7 @@
 ;;
 (define (goops-error format-string . args)
   (save-stack)
-  (scm-error 'goops-error #f format-string args '()))
+  (scm-error 'goops-error #f (oldfmt format-string) args '()))
 
 ;;
 ;; is-a?
@@ -266,7 +267,7 @@
 	  (cond ((not (and (list? exp) (>= (length exp) 2)))
 		 (goops-error "missing or extra expression"))
 		((not (list? (supers exp)))
-		 (goops-error "malformed superclass list: %S" (supers exp)))
+		 (goops-error "malformed superclass list: ~S" (supers exp)))
 		(else
 		 (let ((slot-defs (cons #f '())))
 		   (do ((slots (slots exp) (cdr slots))
@@ -310,10 +311,10 @@
       (let ((tmp1 (find-duplicate supers))
 	    (tmp2 (find-duplicate (map slot-definition-name slots))))
 	(if tmp1
-	    (goops-error "make-class: super class %S is duplicate in class %S"
+	    (goops-error "make-class: super class ~S is duplicate in class ~S"
 			 tmp1 name))
 	(if tmp2
-	    (goops-error "make-class: slot %S is duplicate in class %S"
+	    (goops-error "make-class: slot ~S is duplicate in class ~S"
 			 tmp2 name)))
 
       ;; Everything seems correct, build the class
@@ -333,7 +334,7 @@
     (lambda (exp env)
       (let ((name (cadr exp)))
 	(cond ((not (symbol? name))
-	       (goops-error "bad generic function name: %S" name))
+	       (goops-error "bad generic function name: ~S" name))
 	      ((defined? name env)
 	       `(define ,name
 		  (if (is-a? ,name <generic>)
@@ -363,7 +364,7 @@
     (lambda (exp env)
       (let ((name (cadr exp)))
 	(cond ((not (symbol? name))
-	       (goops-error "bad accessor name: %S" name))
+	       (goops-error "bad accessor name: ~S" name))
 	      ((defined? name env)
 	       `(define ,name
 		  (if (and (is-a? ,name <generic-with-setter>)
@@ -424,10 +425,11 @@
 	(if (and (pair? name)
 		 (eq? (car name) 'setter)
 		 (pair? (cdr name))
+		 (symbol? (cadr name))
 		 (null? (cddr name)))
 	    (let ((name (cadr name)))
 	      (cond ((not (symbol? name))
-		     (goops-error "bad method name: %S" name))
+		     (goops-error "bad method name: ~S" name))
 		    ((defined? name env)
 		     `(begin
 			;; *fixme* Temporary hack for the current module system
@@ -438,8 +440,11 @@
 		     `(begin
 			(define-accessor ,name)
 			(add-method! (setter ,name) (method ,@(cddr exp)))))))
-	    (cond ((not (symbol? name))
-		   (goops-error "bad method name: %S" name))
+	    (cond ((pair? name)
+		   ;; Convert new syntax to old
+		   `(define-method ,(car name) ,(cdr name) ,@(cddr exp)))
+		  ((not (symbol? name))
+		   (goops-error "bad method name: ~S" name))
 		  ((defined? name env)
 		   `(begin
 		      ;; *fixme* Temporary hack for the current module system
@@ -521,10 +526,21 @@
   (slot-set! gf 'methods (compute-new-list-of-methods gf m))
   (let ((specializers (slot-ref m 'specializers)))
     (slot-set! gf 'n-specialized
-	       (max (if (list? specializers)
-			(length specializers)
-			(+ 1 (length* specializers)))
-		    (slot-ref gf 'n-specialized))))
+	       (let ((n-specialized (slot-ref gf 'n-specialized)))
+		 ;; The magnitude indicates # specializers.
+		 ;; A negative value indicates that at least one
+		 ;; method has rest arguments. (Ugly but effective
+		 ;; space optimization saving one slot in GF objects.)
+		 (cond ((negative? n-specialized)
+			(- (max (+ 1 (length* specializers))
+				(abs n-specialized))))
+		       ((list? specializers)
+			(max (length specializers)
+			     n-specialized))
+		       (else
+			(- (+ 1 (max (length* specializers)
+				     n-specialized)))))
+		 )))
   (%invalidate-method-cache! gf)
   (add-method-in-classes! m)
   *unspecified*)
@@ -547,7 +563,7 @@
   (add-method! (primitive-generic-generic pg) m))
 
 (define-method add-method! (obj (m <method>))
-  (goops-error "%S is not a valid generic function" obj))
+  (goops-error "~S is not a valid generic function" obj))
 
 ;;;
 ;;; {Access to meta objects}
@@ -732,19 +748,19 @@
   ((cadr (class-slot-g-n-s class slot)) #f value))
 
 (define-method slot-unbound ((c <class>) (o <object>) s)
-  (goops-error "Slot `%S' is unbound in object %S" s o))
+  (goops-error "Slot `~S' is unbound in object ~S" s o))
 
 (define-method slot-unbound ((c <class>) s)
-  (goops-error "Slot `%S' is unbound in class %S" s c))
+  (goops-error "Slot `~S' is unbound in class ~S" s c))
 
 (define-method slot-unbound ((o <object>))
-  (goops-error "Unbound slot in object %S" o))
+  (goops-error "Unbound slot in object ~S" o))
 
 (define-method slot-missing ((c <class>) (o <object>) s)
-  (goops-error "No slot with name `%S' in object %S" s o))
+  (goops-error "No slot with name `~S' in object ~S" s o))
   
 (define-method slot-missing ((c <class>) s)
-  (goops-error "No class slot with name `%S' in class %S" s c))
+  (goops-error "No class slot with name `~S' in class ~S" s c))
   
 
 (define-method slot-missing ((c <class>) (o <object>) s value)
@@ -753,21 +769,21 @@
 ;;; Methods for the possible error we can encounter when calling a gf
 
 (define-method no-next-method ((gf <generic>) args)
-  (goops-error "No next method when calling %S\nwith arguments %S" gf args))
+  (goops-error "No next method when calling ~S\nwith arguments ~S" gf args))
 
 (define-method no-applicable-method ((gf <generic>) args)
-  (goops-error "No applicable method for %S in call %S"
+  (goops-error "No applicable method for ~S in call ~S"
 	       gf (cons (generic-function-name gf) args)))
 
 (define-method no-method ((gf <generic>) args)
-  (goops-error "No method defined for %S"  gf))
+  (goops-error "No method defined for ~S"  gf))
 
 ;;;
 ;;; {Cloning functions (from rdeline@CS.CMU.EDU)}
 ;;;
 
 (define-method shallow-clone ((self <object>))
-  (let ((clone (%allocate-instance (class-of self)))
+  (let ((clone (%allocate-instance (class-of self) '()))
 	(slots (map slot-definition-name
 		    (class-slots (class-of self)))))
     (for-each (lambda (slot)
@@ -777,7 +793,7 @@
     clone))
 
 (define-method deep-clone  ((self <object>))
-  (let ((clone (%allocate-instance (class-of self)))
+  (let ((clone (%allocate-instance (class-of self) '()))
 	(slots (map slot-definition-name
 		    (class-slots (class-of self)))))
     (for-each (lambda (slot)
@@ -935,7 +951,8 @@
 			    (init-thunk
 			     (standard-get g-n-s))
 			    (else
-			     (bound-check-get g-n-s))))))
+			     (bound-check-get g-n-s)))
+	  #:slot-definition slotdef)))
 
 (define-method compute-setter-method ((class <class>) slotdef)
   (let ((g-n-s (cddr slotdef)))
@@ -943,7 +960,8 @@
           #:specializers (list class <top>)
 	  #:procedure (if (pair? g-n-s)
 			  (cadr g-n-s)
-			  (standard-set g-n-s)))))
+			  (standard-set g-n-s))
+	  #:slot-definition slotdef)))
 
 (define (make-generic-bound-check-getter proc)
   (let ((source (and (closure? proc) (procedure-source proc))))
@@ -951,7 +969,7 @@
 	(let ((obj (caadr source)))
 	  ;; smart closure compilation
 	  (local-eval
-	   `(lambda (,obj) (assert-bound ,(caddr source) ,obj))
+	   `(lambda (,obj) (,assert-bound ,(caddr source) ,obj))
 	   (procedure-environment proc)))
 	(lambda (o) (assert-bound (proc o) o)))))
 
@@ -999,11 +1017,11 @@
 	      (set (cadr l)))
 	  (if (not (and (closure? get)
 			(= (car (procedure-property get 'arity)) 1)))
-	      (goops-error "Bad getter closure for slot `%S' in %S: %S"
+	      (goops-error "Bad getter closure for slot `~S' in ~S: ~S"
 			   slot class get))
 	  (if (not (and (closure? set)
 			(= (car (procedure-property set 'arity)) 2)))
-	    (goops-error "Bad setter closure for slot `%S' in %S: %S"
+	    (goops-error "Bad setter closure for slot `~S' in ~S: ~S"
 			 slot class set)))))
 
   (map (lambda (s)
@@ -1043,17 +1061,38 @@
 
 ;; Support
 
-(define (every test . lists)
-  (let scan ((tails lists))
-    (if (member #t (map null? tails))             ;(any null? lists)
-	#t
-	(and (apply test (map car tails))
-	     (scan (map cdr tails))))))
+(define (only-non-null lst)
+  (filter (lambda (l) (not (null? l))) lst))
 
-(define (collect-if test? list)
-  (cond ((null? list) '())
-	((test? (car list)) (cons (car list) (collect-if test? (cdr list))))
-	(else (collect-if test? (cdr list)))))
+(define (compute-std-cpl c get-direct-supers)
+  (let ((c-direct-supers (get-direct-supers c)))
+    (merge-lists (list c)
+                 (only-non-null (append (map class-precedence-list
+					     c-direct-supers)
+                                        (list c-direct-supers))))))
+
+(define (merge-lists reversed-partial-result inputs)
+  (cond
+   ((every null? inputs)
+    (reverse! reversed-partial-result))
+   (else
+    (let* ((candidate (lambda (c)
+                        (and (not (any (lambda (l)
+                                         (memq c (cdr l)))
+                                       inputs))
+                             c)))
+           (candidate-car (lambda (l)
+                            (and (not (null? l))
+                                 (candidate (car l)))))
+           (next (any candidate-car inputs)))
+      (if (not next)
+          (goops-error "merge-lists: Inconsistent precedence graph"))
+      (let ((remove-next (lambda (l)
+                           (if (eq? (car l) next)
+                               (cdr l)
+                             l))))
+        (merge-lists (cons next reversed-partial-result)
+                     (only-non-null (map remove-next inputs))))))))
 
 ;; Modified from TinyClos:
 ;;
@@ -1065,7 +1104,7 @@
 ;; Mendhekar <anurag@moose.cs.indiana.edu>.
 ;;
 
-(define (compute-std-cpl c get-direct-supers)
+(define (compute-clos-cpl c get-direct-supers)
   (top-sort ((build-transitive-closure get-direct-supers) c)
 	    ((build-constraints get-direct-supers) c)
 	    (std-tie-breaker get-direct-supers)))
@@ -1078,7 +1117,7 @@
     (if (null? elements)
 	result
 	(let ((can-go-in-now
-	       (collect-if
+	       (filter
 		(lambda (x)
 		  (every (lambda (constraint)
 			   (or (not (eq? (cadr constraint) x))
@@ -1092,7 +1131,7 @@
 				(tie-breaker result
 					     can-go-in-now))))
 		(loop
-		 (collect-if (lambda (x) (not (eq? x choice)))
+		 (filter (lambda (x) (not (eq? x choice)))
 			     elements)
 		 constraints
 		 (append result (list choice)))))))))
@@ -1102,7 +1141,7 @@
     (let loop ((pcpl (reverse partial-cpl)))
       (let ((current-elt (car pcpl)))
 	(let ((ds-of-ce (get-supers current-elt)))
-	  (let ((common (collect-if (lambda (x)
+	  (let ((common (filter (lambda (x)
 				      (memq x ds-of-ce))
 				    min-elts)))
 	    (if (null? common)
@@ -1177,7 +1216,7 @@
 	   (set (get-keyword #:slot-set! (slot-definition-options s) #f))
 	   (env (class-environment class)))
        (if (not (and get set))
-	   (goops-error "You must supply a :slot-ref and a :slot-set! in %S"
+	   (goops-error "You must supply a :slot-ref and a :slot-set! in ~S"
 			s))
        (list get set)))
     (else    (next-method))))
@@ -1188,7 +1227,7 @@
 	  (lambda (o v) (set! shared-variable v)))))
 
 (define-method compute-get-n-set ((o <object>) s)
-  (goops-error "Allocation \"%S\" is unknown" (slot-definition-allocation s)))
+  (goops-error "Allocation \"~S\" is unknown" (slot-definition-allocation s)))
 
 (define-method compute-slots ((class <class>))
   (%compute-slots class))
